/*****************************************************************************/
/*
                                 CGIsapi.c

CGIplus-based ISAPI (Internet Server API; Microsoft/Process Software Purveyor
joint specification) extensions (not filters) scripting environment.

The CGIplus environment, due to it's ability to support persistant scripts, and
basic similarities between ISAPI and CGI processing, can quite conveniently
support an ISAPI scripting environment.  This program essentially becomes a
CGIplus wrapper around an ISAPI DLL (dynamic link library, or with VMS known as
a sharable image).

In this environment the ISAPI extension does NOT execute in the server process
space, it executes in a separate subprocess.  Unlike implementations where
ISAPI extensions are loaded into the server space the WASD approach effectively
insulates buggy ISAPI extensions from crashing the server.  The worst they can
do is crash the subprocess.

The down side is the separate process context is more expensive and slower than
the native loadable, but performance measurements indicate persistant CGISAPI
is still approximately four to five times faster than an equivalent CGI script! 
This makes development for the environment worthwhile for those resource-hungry
applications.  Like CGIplus it's persistant.  This means scripts that have high
initialization  overheads (such as opening large databases) have far less
latency on subsequent uses.  A distinct advantage is no multi-threading issues. 
There will only ever be the one thread being executed by CGISAPI!

CGISAPI DLLs must be executed via a CGIplus-enabled path.

CGISAPI successfully loads and executes the Process Software Corporation
Purveyor VMS example extension, SAMPLE_DLL_AXP.DLL and SAMPLE_DLL_VAX.DLL.


CONSIDERATIONS
--------------
This application is designed to be ISAPI 1.0 compliant.  It should also be
vanilla ISAPI 2.0 compliant (not the Microsoft WIN32 variety, so don't think
you'll necessarily be able to grab all those IIS extensions and just recompile
and use ;^)  For experimentation the compliance version check can be set to
allow any version DLL using the /VERSION= qualifier (major digit only).

With CGISAPI multiple instances of any one extension may be active on the one
server (each in an autonomous subprocess, unlike a server-process-space loaded
extension where only one would ever be active at any one time).  Be aware this
could present different concurrency issues than a single multi or single
threaded instance.

When CGIplus subprocesses are idle they can be sys$delprc()ed at any time by
the server at expiry of lifetime or to free up required server resources.  For
this reason ISAPI extensions (scripts) should finalize the processing of
transactions when finished, not leave anything in a state where it's unexpected
demise might corrupt resources or otherwise cause problems (which is fairly
good general advice anyway ;^)  That is, when finished tidy up as much as is
necessary.  Obviously do not exit completely or a complete reactivation will be
necessary (why would you be using something like ISAPI if you didn't know
that?)

CGISAPI is not thread-aware, it doesn't need to be due to the single context of
the subprocess it executes within.  As a consequence, HttpExtensionProc()
cannot be returned from before all processing is complete.  When this function
returns CGISAPI considers the extension is finished and ready to process
another request.  Returning HSE_STATUS_PENDING or using ServerSupportFunction()
with HSE_REQ_DONE_WITH_SESSION are reported as errors and CGISAPI exits.

When CGISAPI invokes the HttpExtensionProc() with the EXTENSION_CONTROL_BLOCK
initialized, any request body is immediately and completely available via the
'cbTotalBytes', 'cbAvailable' and 'lpbData' fields.  No subsequent calls to
ReadClient() are necessary (although can be made).

Any WASD CGI variable can be requested via GetServerVariable(), as well as
ISAPI-specific "ALL_HTTP" and the WASD-specific "*".  Note that some will not
be portable.  If a variable name does not exist '*lpdwSize' is set to zero,
'lpvBuffer' has a null character set at [0] and FALSE is returned.  If the
supplied size overflows 'lpvBuffer' contains a null-terminated string of as
much as would be contained, '*lpdwSize' is set to the required size (including
terminating null) and false is returned.  If found and no overflow '*lpdwSize'
is set to the size of the string (including null) and TRUE is returned.

Output is explicitly set to be buffered by the C RTL for efficiency reasons. 
For occasions where small amounts of output need to be sent to the client
periodically (perhaps as status information during a more extended period of
processing) use WriteClient() with a zero-length 'lpdwBytes' parameter.  This
will fflush() the C RTL buffer.

CGISAPI loaded extensions can exit at any time they wish.  The subprocess
context allows this.  Of course, normally a server-process-space loaded
instance would not be able to do so!

Any status code and log message set in EXTENSION_CONTROL_BLOCK at return from
HttpExtensionProc() are ignored.  Normal request logging is performed by the
server.


DEBUGGING
---------
The CGISAPI implementation of ISAPI includes a facility to assist with
debugging DLLs.  Basic information on function parameters is output to the
client whenever a call is made to an ISAPI function.  This debugging can be
toggled on and off whenever desired using a call to ServerSupportFunction()
with 'dwHSERRequest' parameter set to the WASD values of HSE_REQ_DLL_DEBUG_ON
or HSE_REQ_DLL_DEBUG_OFF.  Once enabled DLL debugging remains active through
multiple uses of a CGISAPI instance, or until disabled, or until the particular
CGISAPI subprocess' lifetime expires.


SERVER CONFIGURATION
--------------------
Ensure the following are in the appropriate sections of HTTPD$CONFIG.

  [DclScriptRunTime]
  .DLL  $CGI-BIN:[000000]CGISAPI.EXE

  [AddType]
  .DLL  application/octet-stream  -  ISAPI extension DLL

Ensure this is in the scripting section of HTTPD$MAP.

  exec+ /isapi/* /cgi-bin/*

DLLs may then be accessed with paths such as:

  http://host.name.domain/isapi/isapiexample.dll


ISAPI RESOURCES
---------------
The code in this application, data structures, etc., has been derived in part
from information contained in the QUE book "Using ISAPI", Stephen Gensua,
et.al., 1997.

  http://www.genusa.com/isapi/
  http://www.microsoft.com/win32dev/apitext/isalegal.htm
  http://www.process.com/news/spec.htm
  http://vms.process.com/

There's not a lot obvious anywhere for anything but Microsoft platforms, let
alone specifically for VMS :^(  Still, the efficiencies of having persistant
scripts, along with ISAPI being a well-documented interface (in contrast to
CGIplus, though which is simpler and more efficient), will provide real
benefits for some sites.


MAINTAINER'S NOTE
-----------------
CgiVar() has been modified specifically for CGISAPI.  It filters the leading
"WWW_" from each record in the CGIPLUSIN stream before storage.  ISAPI only
uses server variables without the VMS-sepcific "WWW_" for CGI symbols so
eliminating processing around this offered an obvious efficiency advantage.
As a consequence, where this application requests specific variables for it's
own purposes the unqualified name must also be used.


PARAMETERS
----------
The full file specification of the DLL must be supplied.


QUALIFIERS
----------
/CHARSET=        "Content-Type: text/html; charset=..."
/DBUG            turns on all "if (Debug)" statements
/VERSION=        set the ISAPI major version digit to allow non-1.0 DLLs
                 (this is for experimentation, it doesn't really change
                  anything, just stops the version check error report)


LOGICAL NAMES
-------------
CGISAPI$DBUG        turns on all "if (Debug)" statements
CGISAPI$DLLDEBUG    turns on DLL debug mode before an extension is loaded
CGISAPI$PARAM       equivalent to (overrides) the command line
                    parameters/qualifiers (define as a system-wide logical)


BUILD DETAILS
-------------
See BUILD_CGISAPI.COM procedure.


COPYRIGHT
---------
Copyright (c) 1999 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY (update SoftwareID as well!)
---------------
14-MAR-1999  MGD  v1.0.0, initial development
*/
/*****************************************************************************/

#ifdef __ALPHA
   char SoftwareID [] = "CGISAPI AXP-1.0.0";
#else
   char SoftwareID [] = "CGISAPI VAX-1.0.0";
#endif

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

/* VMS-related header files */
#include <descrip.h>
#include <libdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "cgisapi.h"

#define boolean int
#define true 1
#define false 0

#ifdef __ALPHA
#   pragma nomember_alignment
#endif

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) (!((x) & STS$M_SUCCESS))
 
#define CGIPLUS_DEBUG 0
/* return NULLs to indicate the CGI variable was not found */
#define CGIPLUS_NONE NULL

#define DEFAULT_CHARSET "ISO-8859-1"

#define CGISAPI_MAJOR_VERSION 1
#define CGISAPI_MINOR_VERSION 0

#define CGI_VARIABLE_PREFIX "WWW_"
#define CGI_VARIABLE_PREFIX_LENGTH sizeof(CGI_VARIABLE_PREFIX)-1
#define HTTP_VARIABLE_PREFIX "HTTP_"
#define HTTP_VARIABLE_PREFIX_LENGTH sizeof(HTTP_VARIABLE_PREFIX)-1

char  Utility [] = "CGISAPI";

boolean  AllowStandardCgi,
         Debug,
         DllDebug,
         IsCgiPlus,
         ThereHasBeenOutput;

int  DllActivatedCount,
     IsapiMajorVersion = CGISAPI_MAJOR_VERSION;

char  *CgiPlusEofPtr,
      *CgiRequestTimeGmtPtr,
      *CgiServerSoftwarePtr,
      *CliCharsetPtr,
      *CliDllFileNamePtr,
      *CharsetPtr;

char  ContentTypeCharset [64];
      
/* required function prototypes */

char* CgiVar (char*);
char* HttpStatusCodeText (int);

FindImageSymbolHandler ();
char* SysGetMsg (int);

BOOL IsapiServerSupportFunction (HCONN, DWORD, LPVOID, LPDWORD, LPDWORD);
BOOL IsapiGetServerVariable (HCONN, LPSTR, LPVOID, LPDWORD);
BOOL IsapiReadClient (HCONN, LPVOID, LPDWORD);
BOOL IsapiWriteClient (HCONN, LPVOID, LPDWORD, DWORD);

/*****************************************************************************/
/*
*/

main ()

{
   /*********/
   /* begin */
   /*********/

   if (getenv ("CGISAPI$DBUG") != NULL) Debug = true;

   if (getenv ("CGISAPI$DLLDEBUG") != NULL) DllDebug = true;

   /* only for testing certain behaviours! */
   if (getenv ("CGISAPI$CGI") != NULL) AllowStandardCgi = true;

   GetParameters ();

   if (CliDllFileNamePtr == NULL)
   {
      fprintf (stdout, "%%%s-E-DLL, not specified\n", Utility);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }

   if (Debug)
      system ("show sym *");
   else
   {
      /* reopen so that '\r' and '\n' are not filtered, only explicit write */
      if ((stdout =
          freopen ("SYS$OUTPUT", "w", stdout, "ctx=bin", "ctx=xplct")) == NULL)
         exit (vaxc$errno);
   }

   /***********/
   /* process */
   /***********/

   IsCgiPlus = ((CgiPlusEofPtr = getenv("CGIPLUSEOF")) != NULL);

   if (!IsCgiPlus && !AllowStandardCgi)
   {
      /* only allows standard CGI environment when debug switched on */
      ErrorGeneral ("ISAPI scripts must be executed in CGIplus environment!",
                    __FILE__, __LINE__);
      exit (SS$_NORMAL);
   }

   do {

      /* if CGIplus, block waiting for next request */
      CgiVar ("");

      if ((CgiServerSoftwarePtr = CgiVar ("SERVER_SOFTWARE")) == NULL)
         CgiServerSoftwarePtr = SoftwareID;

      /* adjust charset according to request environment */
      if ((CharsetPtr = CliCharsetPtr) == NULL)
      {
         if ((CharsetPtr = CgiVar ("REQUEST_CHARSET")) == NULL)
            CharsetPtr = "";
         if (!CharsetPtr[0])
            if ((CharsetPtr = CgiVar ("SERVER_CHARSET")) == NULL)
               CharsetPtr = "";
         if (!CharsetPtr[0]) CharsetPtr = DEFAULT_CHARSET;
      }
      if (CharsetPtr[0])
         sprintf (ContentTypeCharset, "; charset=%s", CharsetPtr);
      else
         ContentTypeCharset[0] = '\0';

      ThereHasBeenOutput = false;

      /* activate the DLL */
      IsapiMain ();

      if (IsCgiPlus)
      {
         /* flush, write end-of-output line, flush */
         fflush (stdout);
         fputs (CgiPlusEofPtr, stdout);
         fflush (stdout);
      }

   } while (IsCgiPlus);

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
This function gets called each time the CGISAPI script gets activated.  The
first call it must load the DLL (shareable image), ensure the two required
entry-points are available and check the ISAPI version of the DLL.  Subsequent
calls it can just invoke the DLL procedure entry-point to execute it's
functionality.
*/

IsapiMain ()

{
   static BOOL (*GetExtensionVersion)(LPHSE_VERSION_INFO);
   static DWORD (*HttpExtensionProc)(LPEXTENSION_CONTROL_BLOCK);
   static $DESCRIPTOR (FileNameDsc, "");
   static $DESCRIPTOR (HttpExtensionProcDsc, "HttpExtensionProc");
   static $DESCRIPTOR (ImageNameDsc, "CGI-BIN:[000000].DLL");
   static $DESCRIPTOR (GetExtensionVersionDsc, "GetExtensionVersion");

   register char  *cptr, *sptr, *zptr;

   boolean  ok;
   int  status,
        MajorVersion,
        MinorVersion;
   char  FileName [256],
         ImageName [256];
   HSE_VERSION_INFO  HseVersion;
   EXTENSION_CONTROL_BLOCK  ExtensionControlBlock;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "IsapiMain() %d\n", DllActivatedCount);

   if (DllActivatedCount++)
   {
      /********************/
      /* subsequent calls */
      /********************/

      IsapiEcbInit (&ExtensionControlBlock);

      /* invoke the DLL main procedure entry-point */
      ok = (*HttpExtensionProc)(&ExtensionControlBlock);
      if (Debug) fprintf (stdout, "(*HttpExtensionProc)() ok:%d\n", ok);

      IsapiEnd (&ExtensionControlBlock, ok);

      return;
   }

   /**************/
   /* first call */
   /**************/

   cptr = CliDllFileNamePtr;

   zptr = (sptr = ImageName) + sizeof(ImageName)-1;
   while (*cptr && *cptr != ']' && sptr < zptr) *sptr++ = *cptr++;
   if (cptr[1] == '[')
   {
      if (sptr < zptr) *sptr++ = *cptr++;
      while (*cptr && *cptr != ']' && sptr < zptr) *sptr++ = *cptr++;
   }
   if (*cptr && sptr < zptr) *sptr++ = *cptr++;
   ImageNameDsc.dsc$a_pointer = sptr;

   zptr = (sptr = FileName) + sizeof(FileName)-1;
   while (*cptr && *cptr != '.' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   FileNameDsc.dsc$a_pointer = FileName;
   FileNameDsc.dsc$w_length = strlen(FileName);
   if (Debug) fprintf (stdout, "|%s|\n", FileName);

   zptr = ImageName + sizeof(ImageName)-1;
   sptr = ImageNameDsc.dsc$a_pointer;
   while (*cptr && *cptr != ';' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   ImageNameDsc.dsc$a_pointer = ImageName;
   ImageNameDsc.dsc$w_length = strlen(ImageName);
   if (Debug) fprintf (stdout, "|%s|\n", ImageName);

   lib$establish (FindImageSymbolHandler);

   status = lib$find_image_symbol (&FileNameDsc, &HttpExtensionProcDsc,
                                   &HttpExtensionProc, &ImageNameDsc);
   if (Debug) fprintf (stdout, "lib$find_image_symbol() %%X%08.08X\n", status);

   lib$revert ();

   if (VMSnok (status))
   {
      /* could not locate the required entry-point in the DLL */
      if (status == LIB$_KEYNOTFOU)
         ErrorGeneral ("DLL HttpExtensionProc() not found",
                       __FILE__, __LINE__);
      else
         ErrorVmsStatus (status, CliDllFileNamePtr, "", __FILE__, __LINE__);
      exit (SS$_NORMAL);
   }

   status = lib$find_image_symbol (&FileNameDsc, &GetExtensionVersionDsc,
                                   &GetExtensionVersion, &ImageNameDsc);
   if (Debug) fprintf (stdout, "lib$find_image_symbol() %%X%08.08X\n", status);

   if (VMSnok (status))
   {
      /* could not locate the required entry-point in the DLL */
      if (status == LIB$_KEYNOTFOU)
         ErrorGeneral ("DLL GetExtensionVersion() not found",
                       __FILE__, __LINE__);
      else
         ErrorVmsStatus (status, CliDllFileNamePtr, "", __FILE__, __LINE__);
      exit (SS$_NORMAL);
   }

   /* invoke the DLL version information entry-point */
   ok = (*GetExtensionVersion)(&HseVersion);
   if (Debug) fprintf (stdout, "(*GetExtensionVersion)() ok:%d\n", ok);

   if (ok)
   {
      /* check version compatibility */
      MajorVersion = HseVersion.dwExtensionVersion >> 16;
      MinorVersion = HseVersion.dwExtensionVersion & 0x0ffff;
      if (Debug)
         fprintf (stdout, "HseVersion: %d.%d |%s|\n",
                  MajorVersion, MinorVersion, HseVersion.lpszExtensionDesc);
      if (MajorVersion > IsapiMajorVersion)
      {
         char  String [256];
         sprintf (String, "DLL version is %d.%d, %s script is %d.%d",
                  MajorVersion, MinorVersion, Utility,
                  IsapiMajorVersion, CGISAPI_MINOR_VERSION);
         ErrorGeneral (String, __FILE__, __LINE__);
         exit (SS$_NORMAL);
      }
   }
   else
   {
      ErrorGeneral ("DLL GetExtensionVersion() failed", __FILE__, __LINE__);
      exit (SS$_NORMAL);
   }

   /********************************/
   /* call the extension procedure */
   /********************************/

   IsapiEcbInit (&ExtensionControlBlock);

   /* invoke the DLL main procedure entry-point */
   ok = (*HttpExtensionProc)(&ExtensionControlBlock);
   if (Debug) fprintf (stdout, "(*HttpExtensionProc)() ok:%d\n", ok);

   IsapiEnd (&ExtensionControlBlock, ok);
}

/*****************************************************************************/
/*
Just continue, to report an error if the image couldn't be activated or the
required symbol not found.
*/

FindImageSymbolHandler ()

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FindImageSymbolHandler()\n");

   return (SS$_CONTINUE);
}

/*****************************************************************************/
/*
Executed after the DLL's main procedure entry-point has returned.
*/

IsapiEnd
(
LPEXTENSION_CONTROL_BLOCK lpEcb,
DWORD HttpExtensionProcReturn
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "IsapiEnd()\n");

   if (Debug)
      fprintf (stdout,
"dwHttpStatusCode: %d\n\
lpszLogData |%s|\n",
               lpEcb->dwHttpStatusCode,
               lpEcb->lpszLogData);

   if (DllDebug)
   {
      fprintf (stdout,
"[HttpExtensionProc() returned: %d\n\
 dwHttpStatusCode: %d\n\
 lpszLogData: |%s|]\n",
         HttpExtensionProcReturn,
         lpEcb->dwHttpStatusCode,
         lpEcb->lpszLogData);
   }

   if (lpEcb->dwHttpStatusCode == HSE_STATUS_PENDING)
   {
      /* we're not threaded here, so anything pending's not supported! */
      ErrorGeneral ("HSE_STATUS_PENDING not supported!",
                    __FILE__, __LINE__);
      exit (SS$_NORMAL);
   }
}

/*****************************************************************************/
/*
Initialize the extension control block, filling out the required fields.
*/

IsapiEcbInit (LPEXTENSION_CONTROL_BLOCK lpEcb)

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "IsapiEcbInit()\n");

   memset (lpEcb, 0, sizeof(EXTENSION_CONTROL_BLOCK));

   lpEcb->cbSize = sizeof(EXTENSION_CONTROL_BLOCK);
   lpEcb->dwVersion = CGISAPI_MAJOR_VERSION << 16 | CGISAPI_MINOR_VERSION;
   lpEcb->ConnID = lpEcb;

   if ((lpEcb->lpszMethod = CgiVar("REQUEST_METHOD")) == NULL)
      lpEcb->lpszMethod = "";
   if ((lpEcb->lpszQueryString = CgiVar("QUERY_STRING")) == NULL)
      lpEcb->lpszQueryString = "";
   if ((lpEcb->lpszPathInfo = CgiVar("PATH_INFO")) == NULL)
      lpEcb->lpszPathInfo = "";
   if ((lpEcb->lpszPathTranslated = CgiVar("PATH_TRANSLATED")) == NULL)
      lpEcb->lpszPathTranslated = "";
   if ((lpEcb->lpszContentType = CgiVar("CONTENT_TYPE")) == NULL)
      lpEcb->lpszContentType = "";

   /* initialize the request body procedure, indicated by a NULL 'lpdBuffer' */
   IsapiReadClient (lpEcb, NULL, NULL);

   lpEcb->GetServerVariable = &IsapiGetServerVariable;
   lpEcb->ReadClient = &IsapiReadClient;
   lpEcb->WriteClient = &IsapiWriteClient;
   lpEcb->ServerSupportFunction = &IsapiServerSupportFunction;

   if (DllDebug)
   {
      fprintf (stdout,
"[%s %s activation: %d]\n",
         SoftwareID, CliDllFileNamePtr, DllActivatedCount);

      if ((cptr = CgiVar("*")) != NULL)
      {
         fprintf (stdout, "[%s", cptr);
         while ((cptr = CgiVar("*")) != NULL)
            fprintf (stdout, "\n %s", cptr);
         fputs ("]\n", stdout);
      }

      fprintf (stdout,
"[HttpExtensionProc()\n\
 ConnID: %d\n\
 lpszMethod: |%s|\n\
 lpszQueryString: |%s|\n\
 lpszPathInfo: |%s|\n\
 lpszPathTranslated: |%s|\n\
 lpszContentType: |%s|\n\
 cbTotalBytes: %d\n\
 cbAvailable: %d\n\
 lpbData: %d]\n",
         lpEcb->ConnID,
         lpEcb->lpszMethod,
         lpEcb->lpszQueryString,
         lpEcb->lpszPathInfo,
         lpEcb->lpszPathTranslated,
         lpEcb->lpszContentType,
         lpEcb->cbTotalBytes,
         lpEcb->cbAvailable,
         lpEcb->lpbData);
   }
}

/*****************************************************************************/
/*
DLL callback that will return the requested CGI-like variable value.  Returns
TRUE if the variable exists, FALSE if it doesn't (sets 'lpvBuffer' to an empty
string).  Returns FALSE if the buffer overflows (puts as much as possible into
it as a null-terminated stream, then sets 'lpdwSize' to what was actually
required).  Special case is "content-length", sets 'lpvBuffer' to a 4 byte
binary value.
*/

BOOL IsapiGetServerVariable
(
HCONN hConn,
LPSTR lpszVariableName,
LPVOID lpvBuffer,
LPDWORD lpdwSize
)
{
   register char  *cptr, *sptr, *zptr;

   int  len;
   char  *CgiVarPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "IsapiGetServerVariable() |%s|\n", lpszVariableName);

   if (DllDebug)
   {
      fprintf (stdout,
"[GetServerVariable()\n\
 hConn: %d\n\
 lpszVariableName: |%s|\n\
 lpvBuffer: %d\n\
 lpdwSize: %d %d\n",
         hConn, lpszVariableName, lpvBuffer,
         lpdwSize, lpdwSize == NULL ? 0 : *lpdwSize);
   }

   if (lpszVariableName[0] == '*' ||
       strsame (lpszVariableName, "ALL_HTTP", -1))
   {
      /************************/
      /* all, or all HTTP_... */
      /************************/

      zptr = (sptr = lpvBuffer) + *lpdwSize;
      while ((cptr = CgiVar("*")) != NULL)
      {
         if (lpszVariableName[0] == '*' ||
             !memcmp (cptr, HTTP_VARIABLE_PREFIX, HTTP_VARIABLE_PREFIX_LENGTH))
         {
            while (*cptr && *cptr != '=' && sptr < zptr) *sptr++ = *cptr++;
            if (*cptr) cptr++;
            if (sptr < zptr) *sptr++ = ':';
            if (sptr < zptr) *sptr++ = ' ';
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = '\n';
            if (sptr >= zptr) break;
         }
      }
      if (sptr >= zptr)
      {
         /* too small, suggest to double it's size */
         *(char*)lpvBuffer = '\0';
         *lpdwSize = *lpdwSize < 1;

         if (DllDebug)
         {
            fprintf (stdout,
" BUFFER OVERFLOW\n\
 lpvBuffer: |%s|\n\
 lpdwsize: %d %d\n\
 return: FALSE]\n",
                lpvBuffer, lpdwSize, lpdwSize == NULL ? 0 : *lpdwSize);
         }

         return (FALSE);
      }
      *sptr = '\0';

      *lpdwSize = sptr - (char*)lpvBuffer + 1;

      if (DllDebug)
      {
         fprintf (stdout,
" FOUND\n\
 lpvBuffer: |%s|\n\
 lpdwsize: %d %d\n\
 return: TRUE]\n",
             lpvBuffer, lpdwSize, lpdwSize == NULL ? 0 : *lpdwSize);
      }

      return (TRUE);
   }

   /***********************/
   /* single CGI variable */
   /***********************/

   /* fudge: Purveyor must use "user-agent" because it's example DLL does! */
   if (strsame (lpszVariableName, "user-agent", -1))
      lpszVariableName = "http_user_agent";

   if ((cptr = CgiVarPtr = CgiVar (lpszVariableName)) == NULL)
   {
      *(char*)lpvBuffer = '\0';
      *lpdwSize = 0;

      if (DllDebug)
      {
         fprintf (stdout,
" NOT FOUND\n\
 lpvBuffer: |%s|\n\
 lpdwsize: %d %d\n\
 return: FALSE]\n",
             lpvBuffer, lpdwSize, lpdwSize == NULL ? 0 : *lpdwSize);
      }

      return (FALSE);
   }
   else
   {
      if (strsame (lpszVariableName, "content_length", -1))
      {
         /* special case, return binary value */

         if (*lpdwSize < sizeof(DWORD))
         {
            *lpdwSize = sizeof(DWORD);

            if (DllDebug)
            {
               fprintf (stdout,
" BUFFER OVERFLOW\n\
 lpvBuffer: |%s|\n\
 lpdwsize: %d %d\n\
 return: FALSE]\n",
                   lpvBuffer, lpdwSize, lpdwSize == NULL ? 0 : *lpdwSize);
            }

            return (FALSE);
         }

         *(DWORD*)lpvBuffer = atoi(cptr);
         *lpdwSize = sizeof(DWORD);

         if (DllDebug)
         {
            fprintf (stdout,
" FOUND\n\
 lpvBuffer: %d %d\n\
 lpdwsize: %d %d\n\
 return: TRUE]\n",
                lpvBuffer, *(DWORD*)lpvBuffer,
                lpdwSize, lpdwSize == NULL ? 0 : *lpdwSize);
         }

         return (TRUE);
      }

      zptr = (sptr = lpvBuffer) + *lpdwSize - 1;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';

      if (sptr < zptr)
      {
         *lpdwSize = sptr - (char*)lpvBuffer + 1;

         if (DllDebug)
         {
            fprintf (stdout,
" FOUND\n\
 lpvBuffer: |%s|\n\
 lpdwsize: %d %d\n\
 return: TRUE]\n",
                lpvBuffer, lpdwSize, lpdwSize == NULL ? 0 : *lpdwSize);
         }

         return (TRUE);
      }
      else
      {
         while (*cptr) cptr++;
         *lpdwSize = cptr - CgiVarPtr + 1;

         if (DllDebug)
         {
            fprintf (stdout,
" BUFFER OVERFLOW\n\
 lpvBuffer: |%s|\n\
 lpdwsize: %d %d\n\
 return: FALSE]\n",
                lpvBuffer, lpdwSize, lpdwSize == NULL ? 0 : *lpdwSize);
         }

         return (FALSE);
      }
   }
}

/*****************************************************************************/
/*
DLL callback to read (further) request content (body) from the client.  CGISAPI
_always_ supplies all the request content with the initial ECB supplied to the
HttpExtensionProc() entry-point, so this really never will need to be called. 
It it is it just indicates there is no more data available.  IsapiEcbInit()
calls this procedure with 'lpvBuffer' set to NULL to initially set the body
content fields correctly whether there is content or not.
*/

BOOL IsapiReadClient
(
HCONN ConnID,
LPVOID lpvBuffer,
LPDWORD lpdwSize
)
{
   static int  BufferCount;
   static char  *BufferPtr;

   char  *cptr;
   EXTENSION_CONTROL_BLOCK  *lpEcb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "IsapiReadClient() %s\n", ConnID);

   lpEcb = (EXTENSION_CONTROL_BLOCK*)ConnID;

   if (lpvBuffer == NULL)
   {
      if ((cptr = CgiVar("CONTENT_LENGTH")) == NULL)
      {
         lpEcb->cbTotalBytes = lpEcb->cbAvailable = BufferCount = 0;
         lpEcb->lpbData = BufferPtr = "";
         return (FALSE);
      }
      else
      {
         lpEcb->cbTotalBytes = atoi(cptr);
         if (Debug)
            fprintf (stdout, "lpEcb->cbTotalBytes: %d\n", lpEcb->cbTotalBytes);
      }

      if (ReadRequestBody (&BufferPtr, &BufferCount))
      {
         if (Debug) fprintf (stdout, "BufferCount: %d\n", BufferCount);
         if (lpEcb->cbTotalBytes == BufferCount)
         {
            /* always return the entire body with the initial call! */
            lpEcb->lpbData = BufferPtr;
            lpEcb->cbAvailable = BufferCount;
            return (TRUE);
         }
         else
         {
            lpEcb->cbTotalBytes = lpEcb->cbAvailable = BufferCount = 0;
            lpEcb->lpbData = BufferPtr = "";
            return (FALSE);
         }
      }
      else
      {
         lpEcb->cbTotalBytes = lpEcb->cbAvailable = BufferCount = 0;
         lpEcb->lpbData = BufferPtr = "";
         return (FALSE);
      }
   }

   if (DllDebug)
   {
      fprintf (stdout,
"[ReadClient()\n\
 ConnID: %d\n\
 lpvBuffer: %d\n\
 lpdwSize: %d %d\n",
         ConnID, lpvBuffer, lpdwSize, lpdwSize == NULL ? 0 : *lpdwSize);
   }

   /* explicit calls to IsapiReadClient() always have nothing to return! */
   *(char*)lpvBuffer = '\0';
   *lpdwSize = 0;

   if (DllDebug)
   {
      fprintf (stdout,
" lpvBuffer: |%s|\n\
 lpdwsize: %d %d\n\
 return: FALSE]\n",
         lpvBuffer, lpdwSize, lpdwSize == NULL ? 0 : *lpdwSize);
   }

   return (FALSE);
}

/*****************************************************************************/
/*
DLL callback, write the supplied buffer to the client.  As output is buffered
by the C RTL, to flush what is buffered at any time call this function with a
zero-length 'lpdwBytes'.
*/

BOOL IsapiWriteClient
(
HCONN ConnID,
LPVOID lpvBuffer,
LPDWORD lpdwBytes,
DWORD dwReserved
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "IsapiWriteClient()\n");

   if (DllDebug)
   {
      fprintf (stdout,
"[WriteClient()\n\
 ConnID: %d\n\
 lpvBuffer: %d\n\
 lpdwSize: %d %d\n\
 dwReserved: %d]\n",
         ConnID, lpvBuffer,
         lpdwBytes, lpdwBytes == NULL ? 0 : *lpdwBytes,
         dwReserved);
   }

   if (*lpdwBytes)
      fwrite (lpvBuffer, *lpdwBytes, 1, stdout);
   else
      fflush (stdout);

   ThereHasBeenOutput = true;

   if (DllDebug) fprintf (stdout, "\n[return: TRUE]\n");

   return (TRUE);
}

/*****************************************************************************/
/*
DLL callback, provide support functions.
*/

BOOL IsapiServerSupportFunction
(
HCONN hConn,
DWORD dwHSERRequest,
LPVOID lpvBuffer,
LPDWORD lpdwSize,
LPDWORD lpdwDataType
)
{
   EXTENSION_CONTROL_BLOCK  *lpEcb;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "IsapiServerSupportFunction() %d %d %d\n",
               dwHSERRequest, lpdwSize == NULL ? 0 : *lpdwSize, lpdwDataType);

   lpEcb = (EXTENSION_CONTROL_BLOCK*)hConn;

   if (DllDebug)
   {
      fprintf (stdout,
"[ServerSupportFunction()\n\
 hConn: %d\n\
 dwHSERRequest: %d\n\
 lpvBuffer: %d\n\
 lpdwSize: %d %d\n\
 lpdwDataType: %d %d\n",
         hConn, dwHSERRequest, lpvBuffer,
         lpdwSize, lpdwSize == NULL ? 0 : *lpdwSize,
         lpdwDataType, lpdwDataType == NULL ? 0 : *lpdwDataType);
   }

   switch (dwHSERRequest)
   {
      case HSE_REQ_SEND_URL_REDIRECT_RESP :

         if (DllDebug || Debug)
            fprintf (stdout, " HSE_REQ_SEND_URL_REDIRECT_RESP]\n");

         fprintf (stdout,
"HTTP/1.0 302 %s\r\n\
Server: %s\r\n\
Date: %s\r\n\
Location: %s\r\n\
\r\n",
                  HttpStatusCodeText(302),
                  CgiServerSoftwarePtr,
                  CgiRequestTimeGmtPtr,
                  lpvBuffer);

         if (DllDebug) fprintf (stdout, "[return: TRUE]\n");

         return (TRUE);

      case HSE_REQ_SEND_URL :

         if (DllDebug || Debug) fprintf (stdout, " HSE_REQ_SEND_URL]\n");

         /* this is in effect a WASD CGI internal redirect */
         fprintf (stdout, "Location: %s\r\n\r\n", lpvBuffer);

         if (DllDebug) fprintf (stdout, "[return: TRUE]\n");

         return (TRUE);

      case HSE_REQ_SEND_RESPONSE_HEADER :

         if (DllDebug || Debug)
            fprintf (stdout, " HSE_REQ_SEND_RESPONSE_HEADER]\n");

         if ((CgiRequestTimeGmtPtr = CgiVar ("REQUEST_TIME_GMT")) == NULL)
            CgiRequestTimeGmtPtr = "?";

         if (lpvBuffer == NULL)
            fprintf (stdout,
"HTTP/1.0 200 %s\r\n\
Server: %s\r\n\
Date: %s\r\n",
                     HttpStatusCodeText(200),
                     CgiServerSoftwarePtr,
                     CgiRequestTimeGmtPtr);
         else
            fputs (lpvBuffer, stdout);

         if (lpdwDataType == NULL)
            fprintf (stdout,
"Content-Type: text/plain%s\r\n\
\r\n",
                     ContentTypeCharset);
         else
            fputs (lpdwDataType, stdout);

         if (DllDebug) fprintf (stdout, "[return: TRUE]\n");

         return (TRUE);

      case HSE_REQ_DONE_WITH_SESSION :

         if (DllDebug || Debug)
            fprintf (stdout, " HSE_REQ_DONE_WITH_SESSION]\n");

         /* we're not threaded here, so anything pending's not supported! */
         ErrorGeneral ("HSE_REQ_DONE_WITH_SESSION not supported!",
                       __FILE__, __LINE__);

         exit (SS$_NORMAL);

      case HSE_REQ_DLL_DEBUG_ON :

         if (DllDebug || Debug)
            fprintf (stdout, " HSE_REQ_DLL_DEBUG_ON\n return: TRUE]\n");

         DllDebug = true;
         return (TRUE);

      case HSE_REQ_DLL_DEBUG_OFF :

         if (DllDebug || Debug)
            fprintf (stdout, " HSE_REQ_DLL_DEBUG_OFF\n return: TRUE]\n");

         DllDebug = false;
         return (TRUE);

      default :

         if (DllDebug) fprintf (stdout, " return: FALSE]\n");

         return (FALSE);
   }
}

/*****************************************************************************/
/*
Read a POSTed request body into a single array of char (doesn't matter whether
it's text or binary) pointed to by 'BufferPtr' with a length of 'BufferCount'.
*/

boolean ReadRequestBody
(
char **BodyPtrPtr,
int *BodyLength
)
{
#define BUFFER_CHUNK 4096

   static int  BufferSize;
   static char  *BufferPtr;

   int  BufferCount,
        ReadCount;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReadRequestBody()\n");

   if (!BufferSize)
   {
      /* reopen output stream so no carriage control gets added to records */
      if ((stdin = freopen ("SYS$INPUT", "r", stdin, "ctx=bin")) == NULL)
         exit (vaxc$errno);
   }

   *BodyPtrPtr = NULL;
   *BodyLength = 0;

   cptr = CgiVar ("REQUEST_METHOD");
   if (!strsame (cptr, "POST", -1)) return (false);

   if (BufferPtr == NULL)
   {
      /* if first call the begin with an initial allocation of space */
      if ((BufferPtr = calloc (1, BUFFER_CHUNK+1)) == NULL)
         exit (vaxc$errno);
      BufferSize = BUFFER_CHUNK;
   }

   BufferCount = 0;

   while (ReadCount =
          fread (BufferPtr+BufferCount, 1, BufferSize-BufferCount, stdin))
   {
      BufferCount += ReadCount;
      if (BufferCount >= BufferSize)
      {
         if ((BufferPtr =
              realloc (BufferPtr, BufferSize+BUFFER_CHUNK+1)) == NULL)
            exit (vaxc$errno);
         BufferSize += BUFFER_CHUNK;
      }
   }
   if (Debug) fprintf (stdout, "size: %d count: %d\n", BufferSize, BufferCount);

   cptr = CgiVar ("CONTENT_TYPE");
   if (strsame (cptr, "application/x-www-form-urlencoded", -1) ||
       strsame (cptr, "text/", 5))
   {
      /* it's text, terminate with a null (not counted in 'BufferCount') */
      BufferPtr[BufferCount] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", BufferPtr);
   }

   *BodyPtrPtr = BufferPtr;
   *BodyLength = BufferCount;
   if (Debug) fprintf (stdout, "%d %d\n", *BodyLength, *BodyPtrPtr);

   return (true);
}

/*****************************************************************************/
/*
Return the a pointer to abbreviated meaning of the supplied HTTP status code.
These are typical of those included on the response header status line.
*/

char* HttpStatusCodeText (int StatusCode)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpStatusCodeText() %d\n", StatusCode);

   switch (StatusCode)
   {
      case 000 : return ("Script Error!");
      case 200 : return ("Success");
      case 201 : return ("Created");
      case 202 : return ("Accepted");
      case 203 : return ("No content");
      case 301 : return ("Moved permanently");
      case 302 : return ("Moved temporarily");
      case 304 : return ("Not modified");
      case 400 : return ("Bad request");
      case 401 : return ("Authorization required");
      case 403 : return ("Forbidden");
      case 404 : return ("Not found");
      case 500 : return ("Internal error");
      case 501 : return ("Not implemented");
      case 502 : return ("Bad gateway");
      case 503 : return ("Service unavailable");
      default :  return ("Unknown code!");
   }
}

/*****************************************************************************/
/*
Get "command-line" parameters, whether from the command-line or from a
configuration symbol or logical containing the equivalent.
*/
                
GetParameters ()

{
   static char  CommandLine [256];
   static unsigned long  Flags = 0;

   register char  *aptr, *cptr, *clptr, *sptr;

   int  status;
   unsigned short  Length;
   char  ch;
   $DESCRIPTOR (CommandLineDsc, CommandLine);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetParameters()\n");

   if ((clptr = getenv ("CGISAPI$PARAM")) == NULL)
   {
      /* get the entire command line following the verb */
      if (VMSnok (status =
          lib$get_foreign (&CommandLineDsc, 0, &Length, &Flags)))
         exit (status);
      (clptr = CommandLine)[Length] = '\0';
   }

   aptr = NULL;
   ch = *clptr;
   for (;;)
   {
      if (aptr != NULL && *aptr == '/') *aptr = '\0';
      if (!ch) break;

      *clptr = ch;
      if (Debug) fprintf (stdout, "clptr |%s|\n", clptr);
      while (*clptr && isspace(*clptr)) *clptr++ = '\0';
      aptr = clptr;
      if (*clptr == '/') clptr++;
      while (*clptr && !isspace (*clptr) && *clptr != '/')
      {
         if (*clptr != '\"')
         {
            clptr++;
            continue;
         }
         cptr = clptr;
         clptr++;
         while (*clptr)
         {
            if (*clptr == '\"')
               if (*(clptr+1) == '\"')
                  clptr++;
               else
                  break;
            *cptr++ = *clptr++;
         }
         *cptr = '\0';
         if (*clptr) clptr++;
      }
      ch = *clptr;
      if (*clptr) *clptr = '\0';
      if (Debug) fprintf (stdout, "aptr |%s|\n", aptr);
      if (!*aptr) continue;

      if (strsame (aptr, "/CHARSET=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         CliCharsetPtr = cptr;
         continue;
      }
      if (strsame (aptr, "/DBUG", -1))
      {
         Debug = true;
         continue;
      }
      if (strsame (aptr, "/VERSION=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         IsapiMajorVersion = atoi(cptr);
         continue;
      }

      if (*aptr == '/')
      {
         fprintf (stdout, "%%%s-E-IVQUAL, unrecognized qualifier\n \\%s\\\n",
                  Utility, aptr+1);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }

      if (CliDllFileNamePtr == NULL)
      {
         CliDllFileNamePtr = aptr;
         continue;
      }

      fprintf (stdout, "%%%s-E-MAXPARM, too many parameters\n \\%s\\\n",
               Utility, aptr);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }
}

/*****************************************************************************/
/*
*/

ErrorGeneral
(
char *Text,
char *SourceFileName,
int SourceLineNumber
)
{
   register char  *cptr;

   /*********/
   /* begin */
   /*********/

   /* 
      The source file format provided by the "__FILE__" macro will be
      "device:[directory]name.type;ver".  Reduce that to "name.type".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (ThereHasBeenOutput)
   {
      fprintf (stdout,
"<P><HR SIZE=3>\n\
<H1>ERROR!</H1>\n\
<!-- %s, %s, %d -->\n\
<P>Reported by server.\n\
<P>%s\n\
</BODY>\n\
</HTML>\n",
      SoftwareID, cptr, SourceLineNumber, Text);
   }
   else
   {
      fprintf (stdout,
"HTTP/1.0 404 Error\r\n\
Server: %s\r\n\
Content-Type: text/html%s\r\n\
\r\n\
<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"module\" CONTENT=\"%s\">\n\
<META NAME=\"line\" CONTENT=\"%d\">\n\
<TITLE>Error 404</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>ERROR!</H1>\n\
<P>Reported by server.\n\
<P>%s\n\
</BODY>\n\
</HTML>\n",
      CgiServerSoftwarePtr, ContentTypeCharset,
      SoftwareID, cptr, SourceLineNumber, Text);
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/

ErrorVmsStatus
(
int StatusValue,
char *Text,
char *HiddenText,
char *SourceFileName,
int SourceLineNumber
)
{
   static char  Message [256];
   static $DESCRIPTOR (MessageDsc, Message);

   register char  *cptr;
   int  status;
   char  *MoreInfoPtr;
   short int  Length;

   /*********/
   /* begin */
   /*********/

   if (VMSok (status = sys$getmsg (StatusValue, &Length, &MessageDsc, 1, 0))) 
   {
      Message[Length] = '\0';
      Message[0] = toupper(Message[0]);
   }
   else
      strcpy (Message, "&quot;sys$getmsg() failed&quot;");

   MoreInfoPtr = "";

   /* 
      The source file format provided by the "__FILE__" macro will be
      "device:[directory]name.type;ver".  Reduce that to "name.type".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (ThereHasBeenOutput)
   {
      fprintf (stdout,
"<P><HR SIZE=3>\n\
<H1>ERROR!</H1>\n\
<!-- %s, %s, %d -->\n\
<P>Reported by server.\n\
<P>%s ... <TT>%s</TT>\n\
<!-- %%X%08.08X \"%s\" -->\n\
%s\
</BODY>\n\
</HTML>\n",
      SoftwareID, cptr, SourceLineNumber,
      Message, Text, StatusValue, HiddenText, MoreInfoPtr);
   }
   else
   {
      fprintf (stdout,
"HTTP/1.0 404 Error\r\n\
Server: %s\r\n\
Content-Type: text/html%s\r\n\
\r\n\
<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"module\" CONTENT=\"%s\">\n\
<META NAME=\"line\" CONTENT=\"%d\">\n\
<TITLE>Error 404</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>ERROR!</H1>\n\
<P>Reported by server.\n\
<P>%s ... <TT>%s</TT>\n\
<!-- %%X%08.08X \"%s\" -->\n\
%s\
</BODY>\n\
</HTML>\n",
      CgiServerSoftwarePtr, ContentTypeCharset,
      SoftwareID, cptr, SourceLineNumber,
      Message, Text, StatusValue, HiddenText, MoreInfoPtr);
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/
 
char* SysGetMsg (int StatusValue)
 
{
   static char  Message [256];
   short int  Length;
   $DESCRIPTOR (MessageDsc, Message);
 
   sys$getmsg (StatusValue, &Length, &MessageDsc, 0, 0);
   Message[Length] = '\0';
   if (Debug) fprintf (stdout, "SysGetMsg() |%s|\n", Message);
   return (Message);
}
 
/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

boolean strsame
(
register char *sptr1,
register char *sptr2,
register int  count
)
{
   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}

/*****************************************************************************/
/*
              ----- MODIFIED SPECIFICALLY FOR CGISAPI! -----

This function provides a simple, uniform mechanism for a C Language script to
access CGI variable values transparently in both the standard CGI and the
CGIplus environments.  Requires the global storage 'CgiPlusEofPtr' to be
non-NULL to indicate a CGIplus environment.

It is completely self-contained. Simply call the function with the CGI variable
name value to be accessed and if that CGI variable exists a pointer to the
value will be returned.  Non-existant variables return CGIPLUS_NONE (see
below).  The values pointed to should NOT be modified in any way (copy the
value if this is required). 

Supplying an empty variable name (i.e. "") will cause the function to block
until the arrival of a request, upon which it just returns a CGIPLUS_NONE
pointer, allowing start-of-request synchronization.  This will also release
allocated memory and reset in preparation for reading the next request, an
alternative to calling with a NULL parameter at the end of request processing
(see below).

For the CGIplus environment supplying a wilcard variable name (i.e. "*")
returns each CGIplus variable string (line from CGIPLUSIN) in turn, with no
more being indicated by a CGIPLUS_NONE pointer.

At the end of a CGIplus request call the function with a NULL parameter to
release allocated memory and reset ready for processing the next request's
variables.  It is not necessary to do this if calling with an empty parameter
to synchronize on the start of a request (see above).

The CGIplus functionality works by initially reading all lines from
CGIPLUSIN until the terminating blank line is encountered. These lines are
stored as a series of contiguous null-terminated strings, terminated by an
empty string, in allocated memory. These strings are searched for the required
variable name and a pointer to its value returned if found.

The value CGIPLUS_NONE shold be set according to the requirements of the
application the function is included within.  If calling functions can handle
NULL pointers being returned then set the value to NULL.  If non-NULL pointers
are prefered then set the value to an empty string.

14-MAR-99  MGD  refine reading of CGIPLUSIN stream to reduce overhead
21-JUN-98  MGD  version 2 of this function attempts to increase efficiency by
                reducing the number of realloc()s required and introducing the
                length of each variable improving search behaviour
01-JUL-97  MGD  version 1
*/

char* CgiVar (char *VarName)

{
#ifndef CGIPLUS_DEBUG
#  define CGIPLUS_DEBUG 0
#endif
#ifndef CGIPLUS_NONE
#   define CGIPLUS_NONE ""
/** #define CGIPLUS_NONE NULL **/
#endif
#define CGIPLUS_LINE_SIZE 1024
#define CGIPLUS_CHUNK 2048

   static int  LoadVariables = 1,
               VarPtrSize = 0;
   static char  *NextPtr = NULL,
                *VarPtr = NULL;

   register char  *cptr, *sptr, *zptr;

   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (CGIPLUS_DEBUG) fprintf (stdout, "CgiVar() |%s|\n", VarName);

   if (CgiPlusEofPtr == NULL)
   {
      /* standard CGI environment, get variable from environment */
      if (!VarName[0]) return (CGIPLUS_NONE);
      if ((cptr = getenv(VarName)) == NULL)
         return (CGIPLUS_NONE);
      else
         return (cptr);
   }

   if (VarName == NULL || !VarName[0])
   {
      /* end of request, reinitialize */
      LoadVariables = 1;
      NextPtr = NULL;
      if (VarName == NULL) return (CGIPLUS_NONE);
   }

   if (LoadVariables)
   {
      /* read lines containing CGIplus variables from <CGIPLUSIN> */
      static FILE  *CgiPlusIn = NULL;

      int  CurrentVarLength;
      char  *LengthPtr;
      char  Line [CGIPLUS_LINE_SIZE];
                                                 
      LoadVariables = 0;

      if (CgiPlusIn == NULL)
         if ((CgiPlusIn = fopen (getenv("CGIPLUSIN"), "r")) == NULL)
            exit (vaxc$errno);

      CurrentVarLength = 0;
      zptr = VarPtr + VarPtrSize;
      while (fgets (Line, sizeof(Line), CgiPlusIn) != NULL)
      {
         /* ignore any non-alphabetic line */
         if (Line[0] != '\n' && !isalpha(Line[0])) continue;
         /* point to the next name=value storage in the variable buffer */
         LengthPtr = VarPtr + CurrentVarLength;
         sptr = LengthPtr + sizeof(unsigned short);
/*
   CGISAPI MODIFICATION: this line replaced by the two following.
   Step over the leading "WWW_", storing only the variable name.
         for (cptr = Line; *cptr; *sptr++ = *cptr++)
*/
         if (*(cptr = Line) != '\n') cptr += CGI_VARIABLE_PREFIX_LENGTH;
         for (/* above */; *cptr; *sptr++ = *cptr++)
         {
            if (sptr < zptr) continue;
            /* out of storage (or first use) get more (or some) memory */
            if ((VarPtr = realloc (VarPtr, VarPtrSize+CGIPLUS_CHUNK)) == NULL)
               exit (vaxc$errno);
            VarPtrSize += CGIPLUS_CHUNK;
            /* recalculate pointers */
            LengthPtr = VarPtr + CurrentVarLength;
            sptr = LengthPtr + sizeof(unsigned short) + (cptr - Line);
            zptr = VarPtr + VarPtrSize;
         }
         /* remove the trailing newline */
         *--sptr = '\0';
         /* insert the length of this name=value pair immediately before it */
         *((unsigned short*)LengthPtr) = Length = cptr - Line;
         /* adjust the value of the current variable storage space used */
         CurrentVarLength += Length + sizeof(unsigned short);
         /* first empty line signals the end of CGIplus variables */
         if (Line[0] == '\n') break;
      }

      if (VarPtr == NULL) return (CGIPLUS_NONE);

      if (CGIPLUS_DEBUG)
      {
         fprintf (stdout, "VarPtr: %d VarPtrSize: %d CurrentVarLength: %d\n",
                  VarPtr, VarPtrSize, CurrentVarLength);
         sptr = VarPtr;
         for (sptr = VarPtr;
              Length = *((unsigned short*)sptr);
              sptr += sizeof(unsigned short) + Length)
            fprintf (stdout, "%3d |%s|\n", Length, sptr+sizeof(unsigned short));
      }
   }

   /* if just waiting for next request return now it's arrived */
   if (!VarName[0]) return (CGIPLUS_NONE);

   if (VarName[0] == '*')
   {
      /* return each stored line one at a time */
      if (NextPtr == NULL) NextPtr = VarPtr;
      if ((Length = *((unsigned short*)(sptr = NextPtr))) == 1)
      {
         NextPtr = NULL;
         if (CGIPLUS_DEBUG) fprintf (stdout, "|%s|=CGIPLUS_NONE\n", VarName);
         return (CGIPLUS_NONE);
      }
      sptr += sizeof(unsigned short);
      NextPtr = sptr + Length;
      if (CGIPLUS_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, sptr);
      return (sptr);
   }

   /* search CGIplus variable strings for required variable */
   for (sptr = VarPtr;
        Length = *((unsigned short*)sptr);
        sptr = zptr + Length)
   {
      /* simple comparison between supplied and in-string variable names */
      zptr = sptr = sptr + sizeof(unsigned short);
      cptr = VarName;
      while (*cptr && *sptr && *sptr != '=')
      {
         if (toupper(*cptr) != toupper(*sptr)) break;
         cptr++;
         sptr++;
      }
      if (*cptr || *sptr != '=') continue;
      /* found, return a pointer to the value */
      if (CGIPLUS_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, sptr+1);
      return (sptr+1);
   }
   /* not found */
   if (CGIPLUS_DEBUG) fprintf (stdout, "|%s|=CGIPLUS_NONE\n", VarName);
   return (CGIPLUS_NONE);
}

/*****************************************************************************/

